/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, OnInit } from "@angular/core";
import { ActivatedRoute, ParamMap, Router } from "@angular/router";
import { Capability } from "../../model/capability";
import { CapabilityRequest } from "../../model/capability-request";
import { CapabilitiesService } from "../../services/capabilities.service";
import { CapabilityRequestService } from "../../services/capability-request.service";
import { CapabilityLauncherService } from "../../services/capability-launcher.service";
import { DataRetrieverService } from "../../services/data-retriever.service";
import { PollingDataUpdaterService } from "../../services/polling-data-updater.service";
import { BehaviorSubject, combineLatest, Observable, Subject } from "rxjs";
import { map, repeatWhen, scan, takeUntil } from "rxjs/operators";
import { Staff } from "../../model/staff";
import {ActiveRequestsService} from "../../services/active-requests.service";
import {StorageService} from "../../../shared/storage/storage.service";

export const defaultSortOrder = "desc"

@Component({
  selector: "app-active-capability-requests",
  templateUrl: "./active-capability-requests.component.html",
  styleUrls: ["./active-capability-requests.component.scss"],
})
export class ActiveCapabilityRequestsComponent implements OnInit {
  public activeCapabilityRequests: Array<CapabilityRequest>;
  public capability: Capability;
  public isPaused: boolean;
  public enabledCapabilities: Array<string>;
  public qaStaff: Array<Staff>;
  public capabilityName: string;
  public stdCalibrationCapability = "std_calibration";
  private activeCapabilityNames: Observable<string>;

  // Observers
  private activeRequestsObserver = {
    next: (request) => {
      this.activeCapabilityRequests = request.active_requests;
    },
    error: (error) => console.error("Error when retrieving list of active capability requests:", error),
  };
  public capabilityObserver = {
    next: (capability) => {
      this.capability = capability;
      this.isPaused = this.capability.paused;
    },
    error: (error) => console.error("Error when retrieving capability:", error),
  };

  // Table sorting
  public sortedActiveRequests$: Observable<any>;
  private sortedColumn$ = new BehaviorSubject<string>("");
  public sortCol = "id";
  public sortDir = this.getSortDirection();

  private sortDirection$ = this.sortedColumn$.pipe(
    scan<string, { col: string; dir: string }>(
      (sort, val) => {
        if (sort.col === val) {
          this.activeRequestsService.saveReqIdSortOrder(this.sortDir === "asc" ? "desc": "asc");
          this.sortDir = this.getSortDirection();
        }

        return { col: val, dir: this.sortDir };
      },
      { dir: this.sortDir, col: "" },
    ),
  );
  public sortOn(column: string) {
    this.sortedColumn$.next(column);
  }

  public getSortDirection(): string {
    return this.activeRequestsService.getSortOrder() ?? defaultSortOrder;
  }

  // to signal when all subscribers should unsubscribe (triggered in tear down)
  private subscribeFlag = new Subject<boolean>();

  constructor(
    private capabilityRequestService: CapabilityRequestService,
    private capabilityLauncherService: CapabilityLauncherService,
    private pollingDataUpdaterService: PollingDataUpdaterService,
    private dataRetriever: DataRetrieverService,
    private capabilitiesService: CapabilitiesService,
    private route: ActivatedRoute,
    private router: Router,
    private storageService: StorageService,
    private activeRequestsService: ActiveRequestsService,
  ) {
    // Initialize table sorting to sort on request ID
    this.sortOn("id");
  }

  ngOnInit(): void {
    this.qaStaff = [];
    // get query parameters: workspaces/active-requests?capability=capabilityName
    this.activeCapabilityNames = this.route.queryParamMap.pipe(map((params: ParamMap) => params.get("capability")));
    this.activeCapabilityNames.subscribe((param) => {
      // if there are no query parameters, set to std_calibration
      this.capabilityName = param ? param : this.stdCalibrationCapability;
    });

    // get list of enabled capabilities
    const enabledCapabilitiesObserver = {
      next: (capabilities) => {
        this.enabledCapabilities = capabilities;
      },
      error: (error) => console.error("Error when retrieving capability version:", error),
    };
    this.capabilitiesService.getEnabledCapabilities().subscribe(enabledCapabilitiesObserver);

    const qaStaffObserver = {
      next: (req) => {
        // this'll distinguish DA and AOD in this.qaStaff
        // like, this.qaStaff["DA"] and this.qaStaff["AOD"]
        if (req["DA"]) {
          this.qaStaff["DA"] = req["DA"];
        } else if (req["AOD"]) {
          this.qaStaff["AOD"] = req["AOD"];
        }
      },
      error: (error) => console.error("Error when retrieving QA staff:", error),
    };

    this.capabilitiesService.getQaStaff("DA").subscribe(qaStaffObserver);
    this.capabilitiesService.getQaStaff("AOD").subscribe(qaStaffObserver);

    // Initial sorting
    this.sortedActiveRequests$ = combineLatest([this.getActiveCapabilityRequests$(), this.sortDirection$]).pipe(
      map(([list, sort]) => (!sort.col ? list : this.sortByColumn(list.active_requests, sort.col, sort.dir))),
    );
  }

  /**
   * Get metadata from capability request parameters if it exists
   * @param request Capability request
   * @return metadata Metadata section of parameters
   */
  public getMetadata(request: CapabilityRequest): any {
    let latestVersion = request.versions[request.versions.length - 1];
    if (latestVersion.parameters.hasOwnProperty("metadata")) {
      return latestVersion.parameters.metadata.valueOf();
    } else {
      return {};
    }
  }

  public togglePauseCapability(): void {
    const togglePauseCapabilityObservable = {
      next: (request) => {},
      error: (error) => {
        console.error("Error when pausing capability: ", error);
      },
    };

    let pauseAction = this.isPaused ? "unpause" : "pause";
    this.dataRetriever
      .togglePauseCapability(this.stdCalibrationCapability, pauseAction)
      .subscribe(togglePauseCapabilityObservable);
  }

  public toggleSRDPCompatibility(request_id: string, version_id: number, isSRDP: boolean): void {
    // stop polling when user clicks checkbox
    this.subscribeFlag.next(false);
    const toggleSRDPCompatibilityObservable = {
      next: (request) => {
        // start polling again after SRDP status has changed
        // TODO: success message to user
        this.subscribeFlag.next(true);
      },
      error: (error) => {
        console.error("Error when toggling SRDP-Compatibility:", error);
        // start polling again with SRDP status unchanged
        // TODO: error message to user
        this.subscribeFlag.next(true);
      },
    };
    this.capabilityRequestService
      .toggleSRDPCompatibility(request_id, version_id, isSRDP)
      .subscribe(toggleSRDPCompatibilityObservable);
  }

  public setSelectedCapability(capabilityName: string): void {
    // set capabilityName from dropdown
    this.capabilityName = capabilityName;

    // Sorting
    this.sortedActiveRequests$ = combineLatest([this.getActiveCapabilityRequests$(), this.sortDirection$]).pipe(
      map(([list, sort]) => (!sort.col ? list : this.sortByColumn(list.active_requests, sort.col, sort.dir))),
    );

    // update query parameters to new capabilityName
    this.router.navigate(["."], {
      relativeTo: this.route,
      queryParams: { capability: this.capabilityName },
    });
  }

  public setStaff(requestId: string, staffMember: Staff): void {
    // stop polling when user opens dropdown
    this.subscribeFlag.next(false);
    const assignQaStaffObservable = {
      next: () => {
        // start polling again when user selects qa staff and req is successful
        // TODO: success message to user
        this.subscribeFlag.next(true);
      },
      error: (error) => {
        console.error("Error when setting" + staffMember.group + ", error: ", error);
        // start polling again when user selects qa staff and req fails
        // TODO: error message to user
        this.subscribeFlag.next(true);
      },
    };
    this.capabilityRequestService.assignQaStaff(requestId, staffMember).subscribe(assignQaStaffObservable);
  }

  // pause polling when user opens qa staff dropdown
  public pausePolling(isDropdownOpen: Event): void {
    isDropdownOpen ? this.subscribeFlag.next(false) : this.subscribeFlag.next(true);
  }

  public clearStaff(requestId: string, group: string): void {
    return this.setStaff(requestId, {
      user_name: "UNASSIGN",
      group: group,
      available: true,
    });
  }

  private getActiveCapabilityRequests$(): Observable<any> {
    return this.pollingDataUpdaterService
      .getDataPoller$(this.capabilityRequestService.getActiveCapabilityRequestsUrl(this.capabilityName))
      .pipe(
        takeUntil(this.subscribeFlag),
        repeatWhen(() => this.subscribeFlag),
      );
  }

  private sortByColumn(list: any[] | undefined, column: string, direction = "desc"): any[] {
    // Set global sort column and direction flags
    this.sortCol = column;
    this.sortDir = direction;

    // Do sort
    return (list || []).sort((a, b) => {
      // Handle observation times (or really any metadata field)
      if (column.includes("obs_")) {
        a = this.getMetadata(a);
        b = this.getMetadata(b);
      }

      // Handle normally accessible members
      if (a[column] > b[column]) {
        return direction === "asc" ? 1 : -1;
      }
      if (a[column] < b[column]) {
        return direction === "asc" ? -1 : 1;
      }
      return 0;
    });
  }
}
